@@ -28,11 +28,10 @@ from datetime import datetime
 
 # Import salt libs
 import salt.config
-import salt.minion
+import salt.cache
 import salt.payload
 import salt.transport
 import salt.loader
-import salt.minion
 import salt.utils
 import salt.utils.args
 import salt.utils.event
@@ -73,7 +72,8 @@ def get_local_client(
         c_path=os.path.join(syspaths.CONFIG_DIR, 'master'),
         mopts=None,
         skip_perm_errors=False,
-        io_loop=None):
+        io_loop=None,
+        auto_reconnect=False):
     '''
     .. versionadded:: 2014.7.0
 
@@ -100,7 +100,8 @@ def get_local_client(
         return LocalClient(
             mopts=opts,
             skip_perm_errors=skip_perm_errors,
-            io_loop=io_loop)
+            io_loop=io_loop,
+            auto_reconnect=auto_reconnect)
 
 
 class LocalClient(object):
@@ -116,6 +117,14 @@ class LocalClient(object):
     running as. (Unless :conf_master:`external_auth` is configured and
     authentication credentials are included in the execution).
 
+    .. note::
+        The LocalClient uses a Tornado IOLoop, this can create issues when
+        using the LocalClient inside an existing IOLoop. If creating the
+        LocalClient in partnership with another IOLoop either create the
+        IOLoop before creating the LocalClient, or when creating the IOLoop
+        use ioloop.current() which will return the ioloop created by
+        LocalClient.
+
     .. code-block:: python
 
         import salt.client
@@ -126,7 +135,7 @@ class LocalClient(object):
     def __init__(self,
                  c_path=os.path.join(syspaths.CONFIG_DIR, 'master'),
                  mopts=None, skip_perm_errors=False,
-                 io_loop=None, keep_loop=False):
+                 io_loop=None, keep_loop=False, auto_reconnect=False):
         '''
         :param IOLoop io_loop: io_loop used for events.
                                Pass in an io_loop if you want asynchronous
@@ -149,6 +158,7 @@ class LocalClient(object):
         self.salt_user = salt.utils.get_specific_user()
         self.skip_perm_errors = skip_perm_errors
         self.key = self.__read_master_key()
+        self.auto_reconnect = auto_reconnect
         self.event = salt.utils.event.get_event(
                 'master',
                 self.opts['sock_dir'],
@@ -171,6 +181,10 @@ class LocalClient(object):
                 key_user = self.opts.get('user', 'root')
         if key_user.startswith('sudo_'):
             key_user = self.opts.get('user', 'root')
+        if salt.utils.is_windows():
+            # The username may contain '\' if it is in Windows
+            # 'DOMAIN\username' format. Fix this for the keyfile path.
+            key_user = key_user.replace('\\', '_')
         keyfile = os.path.join(self.opts['cachedir'],
                                '.{0}_key'.format(key_user))
         # Make sure all key parent directories are accessible
@@ -236,7 +250,7 @@ class LocalClient(object):
         '''
         Common checks on the pub_data data structure returned from running pub
         '''
-        if not pub_data:
+        if pub_data == '':
             # Failed to authenticate, this could be a bunch of things
             raise EauthAuthenticationError(
                 'Failed to authenticate! This is most likely because this '
@@ -246,7 +260,11 @@ class LocalClient(object):
             )
 
         # Failed to connect to the master and send the pub
-        if 'jid' not in pub_data:
+        if 'error' in pub_data:
+            print(pub_data['error'])
+            log.debug('_check_pub_data() error: {0}'.format(pub_data['error']))
+            return {}
+        elif 'jid' not in pub_data:
             return {}
         if pub_data['jid'] == '0':
             print('Failed to connect to the Master, '
@@ -613,7 +631,7 @@ class LocalClient(object):
         was_listening = self.event.cpub
 
         try:
-            pub_data = self.run_job(
+            self.pub_data = self.run_job(
                 tgt,
                 fun,
                 arg,
@@ -623,13 +641,13 @@ class LocalClient(object):
                 listen=True,
                 **kwargs)
 
-            if not pub_data:
-                yield pub_data
+            if not self.pub_data:
+                yield self.pub_data
             else:
                 try:
                     for fn_ret in self.get_cli_event_returns(
-                            pub_data['jid'],
-                            pub_data['minions'],
+                            self.pub_data['jid'],
+                            self.pub_data['minions'],
                             self._get_timeout(timeout),
                             tgt,
                             expr_form,
@@ -642,12 +660,16 @@ class LocalClient(object):
 
                         yield fn_ret
                 except KeyboardInterrupt:
-                    msg = ('Exiting on Ctrl-C\nThis job\'s jid is:\n{0}\n'
-                           'The minions may not have all finished running and any '
-                           'remaining minions will return upon completion. To '
-                           'look up the return data for this job later run:\n'
-                           'salt-run jobs.lookup_jid {0}').format(pub_data['jid'])
-                    raise SystemExit(msg)
+                    raise SystemExit(
+                        '\n'
+                        'This job\'s jid is: {0}\n'
+                        'Exiting gracefully on Ctrl-c\n'
+                        'The minions may not have all finished running and any '
+                        'remaining minions will return upon completion. To look '
+                        'up the return data for this job later, run the following '
+                        'command:\n\n'
+                        'salt-run jobs.lookup_jid {0}'.format(self.pub_data['jid'])
+                    )
         finally:
             if not was_listening:
                 self.event.close_pub()
@@ -887,7 +909,8 @@ class LocalClient(object):
         '''
 
         while True:
-            raw = self.event.get_event(wait=0.01, tag=tag, match_type=match_type, full=True, no_block=True)
+            raw = self.event.get_event(wait=0.01, tag=tag, match_type=match_type, full=True,
+                                       no_block=True, auto_reconnect=self.auto_reconnect)
             yield raw
 
     def get_iter_returns(
@@ -971,6 +994,8 @@ class LocalClient(object):
                         ret[raw['data']['id']]['out'] = raw['data']['out']
                     if 'retcode' in raw['data']:
                         ret[raw['data']['id']]['retcode'] = raw['data']['retcode']
+                    if 'jid' in raw['data']:
+                        ret[raw['data']['id']]['jid'] = raw['data']['jid']
                     if kwargs.get('_cmd_meta', False):
                         ret[raw['data']['id']].update(raw['data'])
                     log.debug('jid {0} return from {1}'.format(jid, raw['data']['id']))
@@ -1003,7 +1028,7 @@ class LocalClient(object):
             # re-do the ping
             if time.time() > timeout_at and minions_running:
                 # since this is a new ping, no one has responded yet
-                jinfo = self.gather_job_info(jid, tgt, tgt_type, **kwargs)
+                jinfo = self.gather_job_info(jid, list(minions - found), 'list', **kwargs)
                 minions_running = False
                 # if we weren't assigned any jid that means the master thinks
                 # we have nothing to send
@@ -1118,7 +1143,7 @@ class LocalClient(object):
         while True:
             time_left = timeout_at - int(time.time())
             wait = max(1, time_left)
-            raw = self.event.get_event(wait, jid)
+            raw = self.event.get_event(wait, jid, auto_reconnect=self.auto_reconnect)
             if raw is not None and 'return' in raw:
                 found.add(raw['id'])
                 ret[raw['id']] = raw['return']
@@ -1271,7 +1296,7 @@ class LocalClient(object):
             # Wait 0 == forever, use a minimum of 1s
             wait = max(1, time_left)
             jid_tag = 'salt/job/{0}'.format(jid)
-            raw = self.event.get_event(wait, jid_tag)
+            raw = self.event.get_event(wait, jid_tag, auto_reconnect=self.auto_reconnect)
             if raw is not None and 'return' in raw:
                 if 'minions' in raw.get('data', {}):
                     minions.update(raw['data']['minions'])
@@ -1352,9 +1377,8 @@ class LocalClient(object):
                 if min_ret.get('failed') is True:
                     if connected_minions is None:
                         connected_minions = salt.utils.minions.CkMinions(self.opts).connected_ids()
-                    minion_cache = os.path.join(self.opts['cachedir'], 'minions', id_, 'data.p')
                     if self.opts['minion_data_cache'] \
-                            and os.path.exists(minion_cache) \
+                            and salt.cache.Cache(self.opts).contains('minions/{0}'.format(id_), 'data') \
                             and connected_minions \
                             and id_ not in connected_minions:
 
@@ -1390,7 +1414,7 @@ class LocalClient(object):
             raise StopIteration()
         # Wait for the hosts to check in
         while True:
-            raw = self.event.get_event(timeout)
+            raw = self.event.get_event(timeout, auto_reconnect=self.auto_reconnect)
             if raw is None or time.time() > timeout_at:
                 # Timeout reached
                 break
@@ -1765,6 +1789,8 @@ class Caller(object):
     '''
     def __init__(self, c_path=os.path.join(syspaths.CONFIG_DIR, 'minion'),
             mopts=None):
+        # Late-import of the minion module to keep the CLI as light as possible
+        import salt.minion
         if mopts:
             self.opts = mopts
         else:
