@@ -5,36 +5,38 @@ A module to wrap (non-Windows) archive calls
 .. versionadded:: 2014.1.0
 '''
 from __future__ import absolute_import
-import os
-import logging
 import contextlib  # For < 2.7 compat
+import errno
+import logging
+import os
+import re
+import shlex
+import tarfile
+import zipfile
+try:
+    from shlex import quote as _quote  # pylint: disable=E0611
+except ImportError:
+    from pipes import quote as _quote
 
 # Import salt libs
 from salt.exceptions import SaltInvocationError, CommandExecutionError
 from salt.ext.six import string_types, integer_types
+from salt.ext.six.moves.urllib.parse import urlparse as _urlparse  # pylint: disable=no-name-in-module
 import salt.utils
+import salt.utils.itertools
 
 # TODO: Check that the passed arguments are correct
 
 # Don't shadow built-in's.
 __func_alias__ = {
-    'zip_': 'zip'
+    'zip_': 'zip',
+    'list_': 'list'
 }
 
 log = logging.getLogger(__name__)
 
 
-HAS_ZIPFILE = False
-try:
-    import zipfile
-    HAS_ZIPFILE = True
-except ImportError:
-    pass
-
-
 def __virtual__():
-    if salt.utils.is_windows():
-        return HAS_ZIPFILE
     commands = ('tar', 'gzip', 'gunzip', 'zip', 'unzip', 'rar', 'unrar')
     # If none of the above commands are in $PATH this module is a no-go
     if not any(salt.utils.which(cmd) for cmd in commands):
@@ -42,6 +44,231 @@ def __virtual__():
     return True
 
 
+def list_(name,
+          archive_format=None,
+          options=None,
+          clean=False,
+          verbose=False,
+          saltenv='base'):
+    '''
+    .. versionadded:: 2016.11.0
+
+    List the files and directories in an tar, zip, or rar archive.
+
+    .. note::
+        This function will only provide results for XZ-compressed archives if
+        xz-utils_ is installed, as Python does not at this time natively
+        support XZ compression in its tarfile_ module.
+
+    name
+        Path/URL of archive
+
+    archive_format
+        Specify the format of the archive (``tar``, ``zip``, or ``rar``). If
+        this argument is omitted, the archive format will be guessed based on
+        the value of the ``name`` parameter.
+
+    options
+        **For tar archives only.** This function will, by default, try to use
+        the tarfile_ module from the Python standard library to get a list of
+        files/directories. If this method fails, then it will fall back to
+        using the shell to decompress the archive to stdout and pipe the
+        results to ``tar -tf -`` to produce a list of filenames. XZ-compressed
+        archives are already supported automatically, but in the event that the
+        tar archive uses a different sort of compression not supported natively
+        by tarfile_, this option can be used to specify a command that will
+        decompress the archive to stdout. For example:
+
+        .. code-block:: bash
+
+            salt minion_id archive.list /path/to/foo.tar.gz options='gzip --decompress --stdout'
+
+        .. note::
+            It is not necessary to manually specify options for gzip'ed
+            archives, as gzip compression is natively supported by tarfile_.
+
+    clean : False
+        Set this value to ``True`` to delete the path referred to by ``name``
+        once the contents have been listed. This option should be used with
+        care.
+
+        .. note::
+            If there is an error listing the archive's contents, the cached
+            file will not be removed, to allow for troubleshooting.
+
+    verbose : False
+        If ``False``, this function will return a list of files/dirs in the
+        archive. If ``True``, it will return a dictionary categorizing the
+        paths into separate keys containing the directory names, file names,
+        and also directories/files present in the top level of the archive.
+
+    saltenv : base
+        Specifies the fileserver environment from which to retrieve
+        ``archive``. This is only applicable when ``archive`` is a file from
+        the ``salt://`` fileserver.
+
+    .. _tarfile: https://docs.python.org/2/library/tarfile.html
+    .. _xz-utils: http://tukaani.org/xz/
+
+    CLI Examples:
+
+    .. code-block:: bash
+
+            salt '*' archive.list /path/to/myfile.tar.gz
+            salt '*' archive.list salt://foo.tar.gz
+            salt '*' archive.list https://domain.tld/myfile.zip
+            salt '*' archive.list ftp://10.1.2.3/foo.rar
+    '''
+    def _list_tar(name, cached, decompress_cmd):
+        try:
+            with contextlib.closing(tarfile.open(cached)) as tar_archive:
+                return [
+                    x.name + '/' if x.isdir() else x.name
+                    for x in tar_archive.getmembers()
+                ]
+        except tarfile.ReadError:
+            if not salt.utils.which('tar'):
+                raise CommandExecutionError('\'tar\' command not available')
+            if decompress_cmd is not None:
+                # Guard against shell injection
+                try:
+                    decompress_cmd = ' '.join(
+                        [_quote(x) for x in shlex.split(decompress_cmd)]
+                    )
+                except AttributeError:
+                    raise CommandExecutionError('Invalid CLI options')
+            else:
+                if salt.utils.which('xz') \
+                        and __salt__['cmd.retcode'](['xz', '-l', cached],
+                                                    python_shell=False,
+                                                    ignore_retcode=True) == 0:
+                    decompress_cmd = 'xz --decompress --stdout'
+
+            if decompress_cmd:
+                cmd = '{0} {1} | tar tf -'.format(decompress_cmd, _quote(cached))
+                result = __salt__['cmd.run_all'](cmd, python_shell=True)
+                if result['retcode'] != 0:
+                    raise CommandExecutionError(
+                        'Failed to decompress {0}'.format(name),
+                        info={'error': result['stderr']}
+                    )
+                ret = []
+                for line in salt.utils.itertools.split(result['stdout'], '\n'):
+                    line = line.strip()
+                    if line:
+                        ret.append(line)
+                return ret
+
+        raise CommandExecutionError(
+            'Unable to list contents of {0}. If this is an XZ-compressed tar '
+            'archive, install xz-utils to enable listing its contents. If it '
+            'is compressed using something other than XZ, it may be necessary '
+            'to specify CLI options to decompress the archive. See the '
+            'documentation for details.'.format(name)
+        )
+
+    def _list_zip(name, cached):
+        # Password-protected ZIP archives can still be listed by zipfile, so
+        # there is no reason to invoke the unzip command.
+        try:
+            with contextlib.closing(zipfile.ZipFile(cached)) as zip_archive:
+                return zip_archive.namelist()
+        except zipfile.BadZipfile:
+            raise CommandExecutionError('{0} is not a ZIP file'.format(name))
+
+    def _list_rar(name, cached):
+        output = __salt__['cmd.run'](
+            ['rar', 'lt', path],
+            python_shell=False,
+            ignore_retcode=False)
+        matches = re.findall(r'Name:\s*([^\n]+)\s*Type:\s*([^\n]+)', output)
+        ret = [x + '/' if y == 'Directory' else x for x, y in matches]
+        if not ret:
+            raise CommandExecutionError(
+                'Failed to decompress {0}'.format(name),
+                info={'error': output}
+            )
+        return ret
+
+    cached = __salt__['cp.cache_file'](name, saltenv)
+    if not cached:
+        raise CommandExecutionError('Failed to cache {0}'.format(name))
+
+    try:
+        parsed = _urlparse(name)
+        path = parsed.path or parsed.netloc
+
+        def _unsupported_format(archive_format):
+            if archive_format is None:
+                raise CommandExecutionError(
+                    'Unable to guess archive format, please pass an '
+                    '\'archive_format\' argument.'
+                )
+            raise CommandExecutionError(
+                'Unsupported archive format \'{0}\''.format(archive_format)
+            )
+
+        if not archive_format:
+            guessed_format = salt.utils.files.guess_archive_type(path)
+            if guessed_format is None:
+                _unsupported_format(archive_format)
+            archive_format = guessed_format
+
+        func = locals().get('_list_' + archive_format)
+        if not hasattr(func, '__call__'):
+            _unsupported_format(archive_format)
+
+        args = (options,) if archive_format == 'tar' else ()
+        try:
+            ret = func(name, cached, *args)
+        except (IOError, OSError) as exc:
+            raise CommandExecutionError(
+                'Failed to list contents of {0}: {1}'.format(
+                    name, exc.__str__()
+                )
+            )
+        except CommandExecutionError as exc:
+            raise
+        except Exception as exc:
+            raise CommandExecutionError(
+                'Uncaught exception \'{0}\' when listing contents of {1}'
+                .format(exc, name)
+            )
+
+        if clean:
+            try:
+                os.remove(cached)
+                log.debug('Cleaned cached archive %s', cached)
+            except OSError as exc:
+                if exc.errno != errno.ENOENT:
+                    log.warning(
+                        'Failed to clean cached archive %s: %s',
+                        cached, exc.__str__()
+                    )
+        if verbose:
+            verbose_ret = {'dirs': [],
+                           'files': [],
+                           'top_level_dirs': [],
+                           'top_level_files': []}
+            for item in ret:
+                if item.endswith('/'):
+                    verbose_ret['dirs'].append(item)
+                    if item.count('/') == 1:
+                        verbose_ret['top_level_dirs'].append(item)
+                else:
+                    verbose_ret['files'].append(item)
+                    if item.count('/') == 0:
+                        verbose_ret['top_level_files'].append(item)
+            ret = verbose_ret
+        return ret
+    except CommandExecutionError as exc:
+        # Reraise with cache path in the error so that the user can examine the
+        # cached archive for troubleshooting purposes.
+        info = exc.info or {}
+        info['archive location'] = cached
+        raise CommandExecutionError(exc.error, info=info)
+
+
 @salt.utils.decorators.which('tar')
 def tar(options, tarfile, sources=None, dest=None,
         cwd=None, template=None, runas=None):
@@ -416,8 +643,14 @@ def zip_(zip_file, sources, template=None, cwd=None, runas=None):
 
 
 @salt.utils.decorators.which('unzip')
-def cmd_unzip(zip_file, dest, excludes=None,
-              template=None, options=None, runas=None, trim_output=False):
+def cmd_unzip(zip_file,
+              dest,
+              excludes=None,
+              options=None,
+              template=None,
+              runas=None,
+              trim_output=False,
+              password=None):
     '''
     .. versionadded:: 2015.5.0
         In versions 2014.7.x and earlier, this function was known as
@@ -444,16 +677,14 @@ def cmd_unzip(zip_file, dest, excludes=None,
 
         .. code-block:: bash
 
-            salt '*' archive.cmd_unzip template=jinja /tmp/zipfile.zip /tmp/{{grains.id}}/ excludes=file_1,file_2
+            salt '*' archive.cmd_unzip template=jinja /tmp/zipfile.zip '/tmp/{{grains.id}}' excludes=file_1,file_2
 
-    options : None
-        Additional command-line options to pass to the ``unzip`` binary.
+    options
+        Optional when using ``zip`` archives, ignored when usign other archives
+        files. This is mostly used to overwrite exsiting files with ``o``.
+        This options are only used when ``unzip`` binary is used.
 
-        .. versionchanged:: 2015.8.0
-
-            The mandatory `-` prefixing has been removed.  An options string
-            beginning with a `--long-option`, would have uncharacteristically
-            needed its first `-` removed under the former scheme.
+        .. versionadded:: 2016.3.1
 
     runas : None
         Unpack the zip file as the specified user. Defaults to the user under
@@ -465,6 +696,23 @@ def cmd_unzip(zip_file, dest, excludes=None,
         The number of files we should output on success before the rest are trimmed, if this is
         set to True then it will default to 100
 
+    password
+        Password to use with password protected zip files
+
+        .. note::
+            This is not considered secure. It is recommended to instead use
+            :py:func:`archive.unzip <salt.modules.archive.unzip>` for
+            password-protected ZIP files. If a password is used here, then the
+            unzip command run to extract the ZIP file will not show up in the
+            minion log like most shell commands Salt runs do. However, the
+            password will still be present in the events logged to the minion
+            log at the ``debug`` log level. If the minion is logging at
+            ``debug`` (or more verbose), then be advised that the password will
+            appear in the log.
+
+        .. versionadded:: 2016.11.0
+
+
     CLI Example:
 
     .. code-block:: bash
@@ -477,23 +725,39 @@ def cmd_unzip(zip_file, dest, excludes=None,
         excludes = [str(excludes)]
 
     cmd = ['unzip']
+    if password:
+        cmd.extend(['-P', password])
     if options:
-        cmd.extend(options.split())
+        cmd.append('{0}'.format(options))
     cmd.extend(['{0}'.format(zip_file), '-d', '{0}'.format(dest)])
 
     if excludes is not None:
         cmd.append('-x')
         cmd.extend(excludes)
-    files = __salt__['cmd.run'](cmd,
-                               template=template,
-                               runas=runas,
-                               python_shell=False).splitlines()
 
-    return _trim_files(files, trim_output)
+    result = __salt__['cmd.run_all'](
+        cmd,
+        template=template,
+        runas=runas,
+        python_shell=False,
+        redirect_stderr=True,
+        output_loglevel='quiet' if password else 'debug')
+
+    if result['retcode'] != 0:
+        raise CommandExecutionError(result['stdout'])
+
+    return _trim_files(result['stdout'].splitlines(), trim_output)
 
 
-def unzip(zip_file, dest, excludes=None,
-          template=None, runas=None, trim_output=False, password=None):
+def unzip(zip_file,
+          dest,
+          excludes=None,
+          options=None,
+          template=None,
+          runas=None,
+          trim_output=False,
+          password=None,
+          extract_perms=True):
     '''
     Uses the ``zipfile`` Python module to unpack zip files
 
@@ -514,6 +778,12 @@ def unzip(zip_file, dest, excludes=None,
         Comma-separated list of files not to unpack. Can also be passed in a
         Python list.
 
+    options
+        This options are only used when ``unzip`` binary is used. In this
+        function is ignored.
+
+        .. versionadded:: 2016.3.1
+
     template : None
         Can be set to 'jinja' or another supported template engine to render
         the command arguments before execution:
@@ -536,23 +806,37 @@ def unzip(zip_file, dest, excludes=None,
 
         salt '*' archive.unzip /tmp/zipfile.zip /home/strongbad/ excludes=file_1,file_2
 
-    password: None
+    password
         Password to use with password protected zip files
 
+        .. note::
+            The password will be present in the events logged to the minion log
+            file at the ``debug`` log level. If the minion is logging at
+            ``debug`` (or more verbose), then be advised that the password will
+            appear in the log.
+
         .. versionadded:: 2016.3.0
 
+    extract_perms : True
+        The Python zipfile_ module does not extract file/directory attributes
+        by default. When this argument is set to ``True``, Salt will attempt to
+        apply the file permision attributes to the extracted files/folders.
+
+        On Windows, only the read-only flag will be extracted as set within the
+        zip file, other attributes (i.e. user/group permissions) are ignored.
+
+        Set this argument to ``False`` to disable this behavior.
+
+        .. versionadded:: 2016.11.0
+
+    .. _zipfile: https://docs.python.org/2/library/zipfile.html
+
     CLI Example:
 
     .. code-block:: bash
 
         salt '*' archive.unzip /tmp/zipfile.zip /home/strongbad/ password='BadPassword'
     '''
-    # https://bugs.python.org/issue15795
-    log.warning('Due to bug 15795 in python\'s zip lib, the permissions of the'
-                ' extracted files may not be preserved when using archive.unzip')
-    log.warning('To preserve the permissions of extracted files, use'
-                ' archive.cmd_unzip')
-
     if not excludes:
         excludes = []
     if runas:
@@ -561,7 +845,7 @@ def unzip(zip_file, dest, excludes=None,
         uinfo = __salt__['user.info'](runas)
         if not uinfo:
             raise SaltInvocationError(
-                'User \'{0}\' does not exist'.format(runas)
+                "User '{0}' does not exist".format(runas)
             )
 
     zip_file, dest = _render_filenames(zip_file, dest, None, template)
@@ -597,6 +881,8 @@ def unzip(zip_file, dest, excludes=None,
                             os.symlink(source, os.path.join(dest, target))
                             continue
                     zfile.extract(target, dest, password)
+                    if extract_perms:
+                        os.chmod(os.path.join(dest, target), zfile.getinfo(target).external_attr >> 16)
     except Exception as exc:
         pass
     finally:
@@ -614,6 +900,65 @@ def unzip(zip_file, dest, excludes=None,
     return _trim_files(cleaned_files, trim_output)
 
 
+def is_encrypted(name, clean=False, saltenv='base'):
+    '''
+    .. versionadded:: 2016.11.0
+
+    Returns ``True`` if the zip archive is password-protected, ``False`` if
+    not. If the specified file is not a ZIP archive, an error will be raised.
+
+    clean : False
+        Set this value to ``True`` to delete the path referred to by ``name``
+        once the contents have been listed. This option should be used with
+        care.
+
+        .. note::
+            If there is an error listing the archive's contents, the cached
+            file will not be removed, to allow for troubleshooting.
+
+
+    CLI Examples:
+
+    .. code-block:: bash
+
+            salt '*' archive.is_encrypted /path/to/myfile.zip
+            salt '*' archive.is_encrypted salt://foo.zip
+            salt '*' archive.is_encrypted https://domain.tld/myfile.zip clean=True
+            salt '*' archive.is_encrypted ftp://10.1.2.3/foo.zip
+    '''
+    cached = __salt__['cp.cache_file'](name, saltenv)
+    if not cached:
+        raise CommandExecutionError('Failed to cache {0}'.format(name))
+
+    archive_info = {'archive location': cached}
+    try:
+        with contextlib.closing(zipfile.ZipFile(cached)) as zip_archive:
+            zip_archive.testzip()
+    except RuntimeError:
+        ret = True
+    except zipfile.BadZipfile:
+        raise CommandExecutionError(
+            '{0} is not a ZIP file'.format(name),
+            info=archive_info
+        )
+    except Exception as exc:
+        raise CommandExecutionError(exc.__str__(), info=archive_info)
+    else:
+        ret = False
+
+    if clean:
+        try:
+            os.remove(cached)
+            log.debug('Cleaned cached archive %s', cached)
+        except OSError as exc:
+            if exc.errno != errno.ENOENT:
+                log.warning(
+                    'Failed to clean cached archive %s: %s',
+                    cached, exc.__str__()
+                )
+    return ret
+
+
 @salt.utils.decorators.which('rar')
 def rar(rarfile, sources, template=None, cwd=None, runas=None):
     '''
