@@ -19,6 +19,13 @@ Module to provide Postgres compatibility to salt.
 
 :note: This module uses MD5 hashing which may not be compliant with certain
     security audits.
+
+:note: When installing postgres from the official postgres repos, on certain
+    linux distributions, either the psql or the initdb binary is *not*
+    automatically placed on the path. Add a configuration to the location
+    of the postgres bin's path to the relevant minion for this module::
+
+        postgres.pg_bin: '/usr/pgsql-9.5/bin/'
 '''
 
 # This pylint error is popping up where there are no colons?
@@ -103,13 +110,31 @@ _PRIVILEGE_TYPE_MAP = {
 
 def __virtual__():
     '''
-    Only load this module if the psql bin exists
+    Only load this module if the psql and initdb bin exist
     '''
-    if all((salt.utils.which('psql'), HAS_CSV)):
-        return True
-    return (False, 'The postgres execution module failed to load: '
-        'either the psql or initdb binary are not in the path or '
-        'the csv library is not available')
+    utils = ['psql', 'initdb']
+    if not HAS_CSV:
+        return False
+    for util in utils:
+        if not salt.utils.which(util):
+            if not _find_pg_binary(util):
+                return (False, '{0} was not found'.format(util))
+    return True
+
+
+def _find_pg_binary(util):
+    '''
+    ... versionadded::  2016.3.2
+
+    Helper function to locate various psql related binaries
+    '''
+    pg_bin_dir = __salt__['config.option']('postgres.bins_dir')
+    util_bin = salt.utils.which(util)
+    if not util_bin:
+        if pg_bin_dir:
+            return os.path.join(pg_bin_dir, util)
+    else:
+        return util_bin
 
 
 def _run_psql(cmd, runas=None, password=None, host=None, port=None, user=None):
@@ -182,9 +207,9 @@ def _run_initdb(name,
 
     if user is None:
         user = runas
-
+    _INITDB_BIN = _find_pg_binary('initdb')
     cmd = [
-        salt.utils.which('initdb'),
+        _INITDB_BIN,
         '--pgdata={0}'.format(name),
         '--username={0}'.format(user),
         '--auth={0}'.format(auth),
@@ -302,8 +327,8 @@ def _psql_cmd(*args, **kwargs):
         kwargs.get('port'),
         kwargs.get('maintenance_db'),
         kwargs.get('password'))
-
-    cmd = [salt.utils.which('psql'),
+    _PSQL_BIN = _find_pg_binary('psql')
+    cmd = [_PSQL_BIN,
            '--no-align',
            '--no-readline',
            '--no-psqlrc',
@@ -457,6 +482,16 @@ def db_exists(name, user=None, host=None, port=None, maintenance_db=None,
     return name in databases
 
 
+# TODO properly implemented escaping
+def _quote_ddl_value(value, quote="'"):
+    if value is None:
+        return None
+    if quote in value:  # detect trivial sqli
+        raise SaltInvocationError(
+            'Unsupported character {0} in value: {1}'.format(quote, value))
+    return "{quote}{value}{quote}".format(quote=quote, value=value)
+
+
 def db_create(name,
               user=None,
               host=None,
@@ -487,16 +522,16 @@ def db_create(name,
     query = 'CREATE DATABASE "{0}"'.format(name)
 
     # "With"-options to create a database
-    with_args = salt.utils.odict.OrderedDict({
+    with_args = salt.utils.odict.OrderedDict([
+        ('TABLESPACE', _quote_ddl_value(tablespace, '"')),
         # owner needs to be enclosed in double quotes so postgres
         # doesn't get thrown by dashes in the name
-        'OWNER': owner and '"{0}"'.format(owner),
-        'TEMPLATE': template,
-        'ENCODING': encoding and '\'{0}\''.format(encoding),
-        'LC_COLLATE': lc_collate and '\'{0}\''.format(lc_collate),
-        'LC_CTYPE': lc_ctype and '\'{0}\''.format(lc_ctype),
-        'TABLESPACE': tablespace,
-    })
+        ('OWNER', _quote_ddl_value(owner, '"')),
+        ('TEMPLATE', template),
+        ('ENCODING', _quote_ddl_value(encoding)),
+        ('LC_COLLATE', _quote_ddl_value(lc_collate)),
+        ('LC_CTYPE', _quote_ddl_value(lc_ctype)),
+    ])
     with_chunks = []
     for key, value in with_args.items():
         if value is not None:
@@ -3053,10 +3088,6 @@ def datadir_init(name,
     runas
         The system user the operation should be performed on behalf of
     '''
-    if salt.utils.which('initdb') is None:
-        log.error('initdb not found in path')
-        return False
-
     if datadir_exists(name):
         log.info('%s already exists', name)
         return False
