@@ -31,19 +31,11 @@ import salt.ext.six as six
 log = logging.getLogger(__name__)
 
 
-def get_pillar(opts, grains, minion_id, saltenv=None, ext=None, env=None, funcs=None,
+def get_pillar(opts, grains, minion_id, saltenv=None, ext=None, funcs=None,
                pillar=None, pillarenv=None):
     '''
     Return the correct pillar driver based on the file_client option
     '''
-    if env is not None:
-        salt.utils.warn_until(
-            'Carbon',
-            'Passing a salt environment should be done using \'saltenv\' '
-            'not \'env\'. This functionality will be removed in Salt Carbon.'
-        )
-        # Backwards compatibility
-        saltenv = env
     ptype = {
         'remote': RemotePillar,
         'local': Pillar
@@ -60,19 +52,11 @@ def get_pillar(opts, grains, minion_id, saltenv=None, ext=None, env=None, funcs=
 
 
 # TODO: migrate everyone to this one!
-def get_async_pillar(opts, grains, minion_id, saltenv=None, ext=None, env=None, funcs=None,
+def get_async_pillar(opts, grains, minion_id, saltenv=None, ext=None, funcs=None,
                pillar=None, pillarenv=None):
     '''
     Return the correct pillar driver based on the file_client option
     '''
-    if env is not None:
-        salt.utils.warn_until(
-            'Carbon',
-            'Passing a salt environment should be done using \'saltenv\' '
-            'not \'env\'. This functionality will be removed in Salt Carbon.'
-        )
-        # Backwards compatibility
-        saltenv = env
     ptype = {
         'remote': AsyncRemotePillar,
         'local': AsyncPillar,
@@ -93,7 +77,8 @@ class AsyncRemotePillar(object):
         self.grains = grains
         self.minion_id = minion_id
         self.channel = salt.transport.client.AsyncReqChannel.factory(opts)
-        self.opts['pillarenv'] = pillarenv
+        if pillarenv is not None or 'pillarenv' not in self.opts:
+            self.opts['pillarenv'] = pillarenv
         self.pillar_override = {}
         if pillar is not None:
             if isinstance(pillar, dict):
@@ -145,7 +130,8 @@ class RemotePillar(object):
         self.grains = grains
         self.minion_id = minion_id
         self.channel = salt.transport.Channel.factory(opts)
-        self.opts['pillarenv'] = pillarenv
+        if pillarenv is not None or 'pillarenv' not in self.opts:
+            self.opts['pillarenv'] = pillarenv
         self.pillar_override = {}
         if pillar is not None:
             if isinstance(pillar, dict):
@@ -328,19 +314,10 @@ class Pillar(object):
             return {}
         return ext
 
-    def __gen_opts(self, opts_in, grains, saltenv=None, ext=None, env=None, pillarenv=None):
+    def __gen_opts(self, opts_in, grains, saltenv=None, ext=None, pillarenv=None):
         '''
         The options need to be altered to conform to the file client
         '''
-        if env is not None:
-            salt.utils.warn_until(
-                'Carbon',
-                'Passing a salt environment should be done using \'saltenv\' '
-                'not \'env\'. This functionality will be removed in Salt '
-                'Carbon.'
-            )
-            # Backwards compatibility
-            saltenv = env
         opts = copy.deepcopy(opts_in)
         opts['file_roots'] = opts['pillar_roots']
         opts['file_client'] = 'local'
@@ -348,7 +325,7 @@ class Pillar(object):
             opts['grains'] = {}
         else:
             opts['grains'] = grains
-        if 'environment' not in opts:
+        if not opts.get('environment'):
             opts['environment'] = saltenv
         opts['id'] = self.minion_id
         if 'pillarenv' not in opts:
@@ -394,8 +371,10 @@ class Pillar(object):
                                 ),
                             self.rend,
                             self.opts['renderer'],
+                            self.opts['renderer_blacklist'],
+                            self.opts['renderer_whitelist'],
                             self.opts['pillarenv'],
-                            _pillar_rend=True
+                            _pillar_rend=True,
                             )
                         ]
             else:
@@ -415,8 +394,10 @@ class Pillar(object):
                                     top,
                                     self.rend,
                                     self.opts['renderer'],
+                                    self.opts['renderer_blacklist'],
+                                    self.opts['renderer_whitelist'],
                                     saltenv=saltenv,
-                                    _pillar_rend=True
+                                    _pillar_rend=True,
                                     )
                                 )
         except Exception as exc:
@@ -453,8 +434,10 @@ class Pillar(object):
                                         ).get('dest', False),
                                     self.rend,
                                     self.opts['renderer'],
+                                    self.opts['renderer_blacklist'],
+                                    self.opts['renderer_whitelist'],
                                     saltenv=saltenv,
-                                    _pillar_rend=True
+                                    _pillar_rend=True,
                                     )
                                 )
                     except Exception as exc:
@@ -607,6 +590,8 @@ class Pillar(object):
             state = compile_template(fn_,
                                      self.rend,
                                      self.opts['renderer'],
+                                     self.opts['renderer_blacklist'],
+                                     self.opts['renderer_whitelist'],
                                      saltenv,
                                      sls,
                                      _pillar_rend=True,
@@ -762,7 +747,7 @@ class Pillar(object):
                 errors.append('The "ext_pillar" option is malformed')
                 log.critical(errors[-1])
                 return {}, errors
-            if run.keys()[0] in self.opts.get('exclude_ext_pillar', []):
+            if next(six.iterkeys(run)) in self.opts.get('exclude_ext_pillar', []):
                 continue
             for key, val in six.iteritems(run):
                 if key not in self.ext_pillars:
@@ -795,16 +780,27 @@ class Pillar(object):
         '''
         top, top_errors = self.get_top()
         if ext:
-            if self.opts.get('ext_pillar_first', False):
+            if self.opts.get('pillar_roots_override_ext_pillar', False) or self.opts.get('ext_pillar_first', False):
+                salt.utils.warn_until('Nitrogen',
+                     'The \'ext_pillar_first\' option has been deprecated and '
+                     'replaced by \'pillar_roots_override_ext_pillar\'.'
+                )
                 self.opts['pillar'], errors = self.ext_pillar({}, pillar_dirs)
                 self.rend = salt.loader.render(self.opts, self.functions)
                 matches = self.top_matches(top)
                 pillar, errors = self.render_pillar(matches, errors=errors)
-                pillar = merge(self.opts['pillar'],
-                               pillar,
-                               self.merge_strategy,
-                               self.opts.get('renderer', 'yaml'),
-                               self.opts.get('pillar_merge_lists', False))
+                if self.opts.get('pillar_roots_override_ext_pillar', False):
+                    pillar = merge(self.opts['pillar'],
+                                   pillar,
+                                   self.merge_strategy,
+                                   self.opts.get('renderer', 'yaml'),
+                                   self.opts.get('pillar_merge_lists', False))
+                else:
+                    pillar = merge(pillar,
+                                   self.opts['pillar'],
+                                   self.merge_strategy,
+                                   self.opts.get('renderer', 'yaml'),
+                                   self.opts.get('pillar_merge_lists', False))
             else:
                 matches = self.top_matches(top)
                 pillar, errors = self.render_pillar(matches)
