@@ -273,6 +273,50 @@ will not run until the specified time has passed. Time should be specified
 in a format supported by the ``dateutil`` library.
 This requires the Python ``dateutil`` library to be installed.
 
+The scheduler also supports ensuring that there are no more than N copies of
+a particular routine running.  Use this for jobs that may be long-running
+and could step on each other or pile up in case of infrastructure outage.
+
+The default for maxrunning is 1.
+
+.. code-block:: yaml
+
+    schedule:
+      long_running_job:
+          function: big_file_transfer
+          jid_include: True
+          maxrunning: 1
+
+By default, data about jobs runs from the Salt scheduler is returned to the
+master.  Setting the ``return_job`` parameter to False will prevent the data
+from being sent back to the Salt master.
+
+.. versionadded:: 2015.5.0
+
+    schedule:
+      job1:
+          function: scheduled_job_function
+          return_job: False
+
+Setting the ``return_job`` parameter to 'nocache' prevents the salt master
+from storing the job in the master cache. Still, an event is fired on the
+master event bus in the form 'salt/job/nocache/ret/myminion'.
+
+It can be useful to include specific data to differentiate a job from other
+jobs.  Using the metadata parameter special values can be associated with
+a scheduled job.  These values are not used in the execution of the job,
+but can be used to search for specific jobs later if combined with the
+return_job parameter.  The metadata parameter must be specified as a
+dictionary, othewise it will be ignored.
+
+.. versionadded:: 2015.5.0
+
+    schedule:
+      job1:
+          function: scheduled_job_function
+          metadata:
+            foo: bar
+
 '''
 
 # Import python libs
@@ -332,7 +376,29 @@ class Schedule(object):
     '''
     Create a Schedule object, pass in the opts and the functions dict to use
     '''
-    def __init__(self, opts, functions, returners=None, intervals=None):
+    instance = None
+
+    def __new__(cls, opts, functions, returners=None, intervals=None, cleanup=None):
+        '''
+        Only create one instance of Schedule
+        '''
+        if cls.instance is None:
+            log.debug('Initializing new Schedule')
+            # we need to make a local variable for this, as we are going to store
+            # it in a WeakValueDictionary-- which will remove the item if no one
+            # references it-- this forces a reference while we return to the caller
+            cls.instance = object.__new__(cls)
+            cls.instance.__singleton_init__(opts, functions, returners, intervals, cleanup)
+        else:
+            log.debug('Re-using Schedule')
+        return cls.instance
+
+    # has to remain empty for singletons, since __init__ will *always* be called
+    def __init__(self, opts, functions, returners=None, intervals=None, cleanup=None):
+        pass
+
+    # an init for the singleton instance to call
+    def __singleton_init__(self, opts, functions, returners=None, intervals=None, cleanup=None):
         self.opts = opts
         self.functions = functions
         if isinstance(intervals, dict):
@@ -348,6 +414,12 @@ class Schedule(object):
         # Keep track of the lowest loop interval needed in this variable
         self.loop_interval = six.MAXSIZE
         clean_proc_dir(opts)
+        if cleanup:
+            for prefix in cleanup:
+                self.delete_job_prefix(prefix)
+
+    def __getnewargs__(self):
+        return self.opts, self.functions, self.returners, self.intervals, None
 
     def option(self, opt):
         '''
@@ -379,7 +451,11 @@ class Schedule(object):
         log.debug('Persisting schedule')
         try:
             with salt.utils.fopen(schedule_conf, 'wb+') as fp_:
-                fp_.write(yaml.dump({'schedule': self.opts['schedule']}))
+                fp_.write(
+                    salt.utils.to_bytes(
+                        yaml.dump({'schedule': self.option('schedule')})
+                    )
+                )
         except (IOError, OSError):
             log.error('Failed to persist the updated schedule',
                       exc_info_on_loglevel=logging.DEBUG)
@@ -390,16 +466,16 @@ class Schedule(object):
         '''
         if where is None or where != 'pillar':
             # ensure job exists, then delete it
-            if name in self.opts['schedule']:
-                del self.opts['schedule'][name]
-            schedule = self.opts['schedule']
+            schedule = self.option('schedule')
+            if name in schedule:
+                del schedule[name]
         else:
             # If job is in pillar, delete it there too
             if 'schedule' in self.opts['pillar']:
                 if name in self.opts['pillar']['schedule']:
                     del self.opts['pillar']['schedule'][name]
             schedule = self.opts['pillar']['schedule']
-            log.warn('Pillar schedule deleted. Pillar refresh recommended. Run saltutil.refresh_pillar.')
+            log.warning('Pillar schedule deleted. Pillar refresh recommended. Run saltutil.refresh_pillar.')
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
@@ -413,6 +489,38 @@ class Schedule(object):
         if persist:
             self.persist()
 
+    def delete_job_prefix(self, name, persist=True, where=None):
+        '''
+        Deletes a job from the scheduler.
+        '''
+        if where is None or where != 'pillar':
+            # ensure job exists, then delete it
+            schedule = self.option('schedule')
+            for job in list(schedule.keys()):
+                if job.startswith(name):
+                    del schedule[job]
+        else:
+            # If job is in pillar, delete it there too
+            if 'schedule' in self.opts['pillar']:
+                for job in list(self.opts['pillar']['schedule'].keys()):
+                    if job.startswith(name):
+                        del self.opts['pillar']['schedule'][job]
+            schedule = self.opts['pillar']['schedule']
+            log.warning('Pillar schedule deleted. Pillar refresh recommended. Run saltutil.refresh_pillar.')
+
+        # Fire the complete event back along with updated list of schedule
+        evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
+        evt.fire_event({'complete': True, 'schedule': schedule},
+                       tag='/salt/minion/minion_schedule_delete_complete')
+
+        # remove from self.intervals
+        for job in list(self.intervals.keys()):
+            if job.startswith(name):
+                del self.intervals[job]
+
+        if persist:
+            self.persist()
+
     def add_job(self, data, persist=True):
         '''
         Adds a new job to the scheduler. The format is the same as required in
@@ -435,17 +543,18 @@ class Schedule(object):
 
         new_job = next(six.iterkeys(data))
 
-        if new_job in self.opts['schedule']:
+        schedule = self.option('schedule')
+        if new_job in schedule:
             log.info('Updating job settings for scheduled '
                      'job: {0}'.format(new_job))
         else:
             log.info('Added new job {0} to scheduler'.format(new_job))
 
-        self.opts['schedule'].update(data)
+        schedule.update(data)
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
-        evt.fire_event({'complete': True, 'schedule': self.opts['schedule']},
+        evt.fire_event({'complete': True, 'schedule': schedule},
                        tag='/salt/minion/minion_schedule_add_complete')
 
         if persist:
@@ -459,8 +568,8 @@ class Schedule(object):
             self.opts['pillar']['schedule'][name]['enabled'] = True
             schedule = self.opts['pillar']['schedule']
         else:
-            self.opts['schedule'][name]['enabled'] = True
-            schedule = self.opts['schedule']
+            schedule = self.option('schedule')
+            schedule[name]['enabled'] = True
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
@@ -480,8 +589,8 @@ class Schedule(object):
             self.opts['pillar']['schedule'][name]['enabled'] = False
             schedule = self.opts['pillar']['schedule']
         else:
-            self.opts['schedule'][name]['enabled'] = False
-            schedule = self.opts['schedule']
+            schedule = self.option('schedule')
+            schedule[name]['enabled'] = False
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
@@ -502,9 +611,9 @@ class Schedule(object):
                 self.delete_job(name, persist, where=where)
             self.opts['pillar']['schedule'][name] = schedule
         else:
-            if name in self.opts['schedule']:
+            if name in self.option('schedule'):
                 self.delete_job(name, persist, where=where)
-            self.opts['schedule'][name] = schedule
+            self.option('schedule')[name] = schedule
 
         if persist:
             self.persist()
@@ -513,7 +622,7 @@ class Schedule(object):
         '''
         Run a schedule job now
         '''
-        schedule = self.opts['schedule']
+        schedule = self.option('schedule')
         if 'schedule' in self.opts['pillar']:
             schedule.update(self.opts['pillar']['schedule'])
         data = schedule[name]
@@ -545,54 +654,51 @@ class Schedule(object):
         else:
             thread_cls = threading.Thread
 
-        proc = thread_cls(target=self.handle_func, args=(multiprocessing_enabled, func, data))
         if multiprocessing_enabled:
             with default_signals(signal.SIGINT, signal.SIGTERM):
+                proc = thread_cls(target=self.handle_func, args=(multiprocessing_enabled, func, data))
                 # Reset current signals before starting the process in
                 # order not to inherit the current signal handlers
                 proc.start()
-        else:
-            proc.start()
-        if multiprocessing_enabled:
             proc.join()
+        else:
+            proc = thread_cls(target=self.handle_func, args=(multiprocessing_enabled, func, data))
+            proc.start()
 
     def enable_schedule(self):
         '''
         Enable the scheduler.
         '''
-        self.opts['schedule']['enabled'] = True
+        schedule = self.option('schedule')
+        schedule['enabled'] = True
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
-        evt.fire_event({'complete': True, 'schedule': self.opts['schedule']},
+        evt.fire_event({'complete': True, 'schedule': schedule},
                        tag='/salt/minion/minion_schedule_enabled_complete')
 
     def disable_schedule(self):
         '''
         Disable the scheduler.
         '''
-        self.opts['schedule']['enabled'] = False
+        schedule = self.option('schedule')
+        schedule['enabled'] = False
 
         # Fire the complete event back along with updated list of schedule
         evt = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
-        evt.fire_event({'complete': True, 'schedule': self.opts['schedule']},
+        evt.fire_event({'complete': True, 'schedule': schedule},
                        tag='/salt/minion/minion_schedule_disabled_complete')
 
     def reload(self, schedule):
         '''
         Reload the schedule from saved schedule file.
         '''
-
         # Remove all jobs from self.intervals
         self.intervals = {}
 
-        if 'schedule' in self.opts:
-            if 'schedule' in schedule:
-                self.opts['schedule'].update(schedule['schedule'])
-            else:
-                self.opts['schedule'].update(schedule)
-        else:
-            self.opts['schedule'] = schedule
+        if 'schedule' in schedule:
+            schedule = schedule['schedule']
+        self.opts.setdefault('schedule', {}).update(schedule)
 
     def list(self, where):
         '''
@@ -603,9 +709,9 @@ class Schedule(object):
             if 'schedule' in self.opts['pillar']:
                 schedule.update(self.opts['pillar']['schedule'])
         elif where == 'opts':
-            schedule.update(self.opts['schedule'])
+            schedule.update(self.option('schedule'))
         else:
-            schedule.update(self.opts['schedule'])
+            schedule.update(self.option('schedule'))
             if 'schedule' in self.opts['pillar']:
                 schedule.update(self.opts['pillar']['schedule'])
 
@@ -776,6 +882,9 @@ class Schedule(object):
                         # Send back to master so the job is included in the job list
                         mret = ret.copy()
                         mret['jid'] = 'req'
+                        if data.get('return_job') == 'nocache':
+                            # overwrite 'req' to signal to master that this job shouldn't be stored
+                            mret['jid'] = 'nocache'
                         event = salt.utils.event.get_event('minion', opts=self.opts, listen=False)
                         load = {'cmd': '_return', 'id': self.opts['id']}
                         for key, value in six.iteritems(mret):
